#! /usr/bin/python
# -*- coding: utf-8 -*-
"""
# merge_svgs.py
Merge SVG files to create poster file
This program changes the page size  

Copyright (C) July 2018 George Zhang

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import inkex
import simplestyle, simpletransform
import sys
import math
import copy
import os
import re

# this function was modified from the composeParents function in simpletransform
# the original function has a error, it needs to test the node.getparent() in 
#  the second if test.  
# this function could be further improved for corner cases. 
def composeParents_Mod(node, mat):
    trans = node.get('transform')
    if trans:
        mat = simpletransform.composeTransform(simpletransform.parseTransform(trans), mat)
    if node.getparent() is not None and node.getparent().tag == inkex.addNS('g','svg'):
        mat = composeParents_Mod(node.getparent(), mat)
    return mat


class Merge_Svgs(inkex.Effect):
  def __init__(self):
    inkex.Effect.__init__(self)
    self.OptionParser.add_option("--isFirstRun", action="store", type="inkbool", dest="bFirstRun", default=False )
    self.OptionParser.add_option("--lineNumber", action="store", type="int", dest="lineN", default="1" )
    self.OptionParser.add_option("--placeLineNumber", action="store", type="int", dest="placeLN", default="1" )


  def effect(self):
    PAGE_W = 24
    PAGE_H = 36
    
    w = PAGE_W
    h = PAGE_H
    if self.options.bFirstRun: 
      self.change_SVG_size(w, h)
      bk_layer = self.find_create_layer(self.document.getroot(), 'bk_layer')
      self.draw_lines(bk_layer, w, h)
    else: 
      
      textpath = 'sign_list.txt'
      filepath = 'sign_svgs/'
      
      line_start = self.options.lineN -1
      place_start = self.options.placeLN - 1
      
      fnlist = self.fn_list(textpath)
      if line_start >= 0 and line_start < len(fnlist):
        fnlist = fnlist[line_start:]
      
      for ii, name in enumerate(fnlist):
        
        if name == '//':
          break
        i = ii % (w-2)
        j = ii // (w-2) + place_start
        p = self.real_path(name, filepath)
        self.place_sign(p, i, j)
        self.place_text(name, i, j)
      
    # p1 = self.real_path('R1-1', filepath)
    # self.place_sign(p1, 0, 0)
    
  def place_text(self, name, i, j):
    x, y = self.text_location(i, j)
    layer = self.find_create_layer(self.document.getroot(), 'text_layer')
    elem = self.create_text(x, y, name)
    layer.append(elem)
  
  def create_text(self, x, y, name):
    
    style_d = {'font-size': '4.233px',  # 12 pt
             'font-family': 'Roboto', 
         #    'text-align': 'center',
             'text-anchor': 'middle',
             'fill': '#000000',
             'stroke': 'none'
            }
    t = inkex.etree.Element('text')
    t.set(inkex.addNS('space', 'xml'), 'preserve')
    t.set('x', str(x))
    t.set('y', str(y))
    t.set('style', simplestyle.formatStyle(style_d))
    
    sp = inkex.etree.SubElement(t, 'tspan')
    sp.set(inkex.addNS('role', 'sodipodi'), 'line')
    sp.text = name 
    return t
    
    # create an element and add it to a new layer
    # fmt = <text xml:space="preserve"
            # x="44.130253" y="-637.38019"
            # style="font-style:normal;font-variant:normal;font-weight:normal;
            # font-stretch:normal;font-size:4.233px;line-height:1.25;
            # font-family:Roboto;-inkscape-font-specification:'Roboto, Normal';
            # font-variant-ligatures:normal;font-variant-caps:normal;
            # font-variant-numeric:normal;font-feature-settings:normal;
            # text-align:center;letter-spacing:0px;word-spacing:0px;
            # writing-mode:lr;text-anchor:middle;fill:#000000;fill-opacity:1;
            # stroke:none;stroke-width:0.25509986;"
          # >
            # <tspan sodipodi:role="line" x="44.130253" y="-637.38019">W11-11</tspan>
          # </text>
    
  def real_path(self, name, filepath):
    
    file_list = self.file_list(filepath)
    name_mod = name[0] + '0' + name[1:] 
    name_mod2 = name[0] + '0' +name[1:3] + '0' + name[3:]
    name_mod3 = name[0:4]+'0'+name[4:] # this code can be optimized
        
    for fn in file_list:
      r3 = name.lower()+' ' == fn.lower()[:len(name)+1]
      r2 = name_mod2.lower()+' ' == fn.lower()[:len(name_mod2)+1]
      r1 = name_mod.lower()+' ' == fn.lower()[:len(name_mod)+1]
      r0 = name_mod3.lower()+' ' == fn.lower()[:len(name_mod3)+1]
      
      if name.lower() == fn.lower() or r3 or r2 or r1 or r0:
        return os.path.join(filepath, fn) + '.svg'
    return os.path.join(filepath, 'R1-1.svg')
    
    
  def fn_list(self, path):
    f = open(path) 
    fn_l = [line.rstrip() for line in f.readlines()]
    f.close()
    return fn_l
    
    
  def file_list(self, path):
    fileList = []
    for f in os.listdir(path):
      if f.endswith('.svg'):
        fileList.append(f[:-4])
    return fileList
    
    
  def place_sign(self, path, i, j):
    layer = self.read_svg(path)
    try:
      self.locate_sign(layer, self.box_dimension(i, j))
    except Exception:
      inkex.debug(path)
    sign_layer = self.find_create_layer(self.document.getroot(), 'sign_layer')
    sign_layer.append(layer)

  def change_SVG_size(self, width, height):
    ratio = 25.4
    svg_elem = self.document.getroot()
    
    page_width = width
    page_height = height
    
    svg_elem.set('width', str(page_width) + 'in')
    svg_elem.set('height', str(page_height) + 'in')
    svg_elem.set('viewBox', '0 0 ' + str(page_width * ratio) + ' ' 
                   + str(page_height * ratio) )  
                   
  def draw_SVG_line(self, (x1, y1), (x2, y2), style, name, parent):
    '''style is a dict'''
    line_attribs = {'style': simplestyle.formatStyle(style), 
              inkex.addNS('label', 'inkscape'): name,
              'd': 'M ' + str(x1) + ',' + str(y1) + ' L' +
                 str(x2) + ',' + str(y2) }
    elm = inkex.etree.SubElement(parent, inkex.addNS('path', 'svg'), line_attribs )
    return elm
                   
  def draw_lines(self, layer, width, height):
    line_style = {'stroke': '#000000',
         'stroke-width': str(self.unittouu('1px')), 
         'fill': 'none' }
    w = width
    h = height
    for i in range(2, h):
      y_str = str(i) + 'in'
      x1, x2, y1, y2 = map(self.unittouu, ['1in', str(w-1)+'in', y_str, y_str])
      self.draw_SVG_line( (x1, y1), (x2, y2), line_style, 'line', layer)
      
    for j in range(1, w):
      x_str = str(j) + 'in'
      x1, x2, y1, y2 = ( self.unittouu(x) for x in [x_str, x_str, '2in', str(h-1)+'in'] )
      self.draw_SVG_line( (x1, y1), (x2, y2), line_style, 'line', layer)
      
  def grid_center(self, i, j):
    ''' i in [0, 34), y in [0, 45)''' 
    ratio = 25.4
    x = (i + 1 + 0.5) * ratio
    y = (j + 2 + 0.5) * ratio
    return (x, y)
    
  def box_dimension(self, i, j):
    ''' returns x, y, width, height'''
    ratio = 25.4
    SIZE = 0.75
    (xc, yc)  = self.grid_center(i, j)
    x = xc - SIZE / 2.0 * ratio
    y = yc - (1.0 / 2.0 - 0.05) * ratio
    return (x, y, SIZE * ratio, SIZE * ratio)
    
  def text_location(self, i, j):
    ratio = 25.4
    SIZE = 0.75
    (xc, yc)  = self.grid_center(i, j)
    x = xc
    y = yc + (1 - SIZE) / 2.0 * 3.0 * ratio
    y = y + 0.14 / 2.0 * ratio # manual adjustment
    return (x, y)
      
  def read_svg(self, path):
    e = inkex.Effect()
    arglist = [ ]
    arglist.append(path)
    filename = os.path.basename(path)
    fn, ext = os.path.splitext(filename)
  
    e.affect(args = arglist, output=False)
    
    elem = e.document.getroot()
    width = e.getDocumentWidth()
    height = e.getDocumentHeight()
    
    xpath = '//svg:g[@inkscape:groupmode="layer"]'
    xpath += '|//g[@inkscape:groupmode="layer"]'
    
    elem_layers = e.document.xpath(xpath, namespaces=inkex.NSS)
                   
                   
    if elem_layers:         # old style files
      if len(elem_layers) == 1 : # one layer, 2004 style, change to group
        elem_layer = elem_layers[0]
        attr = "{%s}%s" % (inkex.NSS['inkscape'], 'label') 
          # or inkex.addNS('label', 'inkscape') 
        elem_layer.set(attr, fn)
        gr_attr = "{%s}%s" % (inkex.NSS['inkscape'], 'groupmode')
        del elem_layer.attrib[gr_attr]  # remove an attribute of element
                                        # it becomes a group element
        del elem_layer.attrib['transform']
        del elem_layer.attrib['id'] # inkscape assign one
        gp = elem_layer
      else:   # more than one layer, group them together
        gp = self.create_group(fn)
        for elem_layer in elem_layers:
          gr_attr = "{%s}%s" % (inkex.NSS['inkscape'], 'groupmode')
          del elem_layer.attrib[gr_attr]  
          #del elem_layer.attrib['transform']
          del elem_layer.attrib['id'] # inkscape assign one
          gp.append(elem_layer)
          
    else:
      # no layer 2012 style, group, no layer
      gp_list = e.document.xpath('/svg/g|/svg:svg/svg:g', namespaces=inkex.NSS)
      if gp_list:
        gp = gp_list[0]
      else:
        gp = self.create_group(fn)  # handle error svg files
        svg = e.document.xpath('/svg|/svg:svg', namespaces=inkex.NSS)
        for node in svg[0].iterchildren():
          if node.get('d'):
            gp.append(node)
        #inkex.debug('path: ' + path)

      #pass
    elem_ret = copy.deepcopy(gp)
    return elem_ret

    
  def compute_signbox(self, elem):
    bbox = None
    tag_list = ['rect', inkex.addNS('rect', 'svg'), 
                'line', inkex.addNS('line', 'svg'), 
                'circle', inkex.addNS('circle', 'svg'), 
                'ellipse', inkex.addNS('ellipse', 'svg')]
                
    for node in elem.iterdescendants():
      if node.get('d') or node.get('points') or node.tag in tag_list:
        mat = [[1, 0, 0], [0, 1, 0]] #simpletransform.parseTransform(node.get('transform')) # double count
        m = composeParents_Mod(node, mat)
        node_c = copy.deepcopy(node)
        node_c.set('transform', simpletransform.formatTransform(m))  # apply changes to a copy
        bbox = simpletransform.boxunion( simpletransform.computeBBox([node_c, ]),bbox)

    return bbox
    
    
  def locate_sign(self, elem, box_dimension):
    bbox = self.compute_signbox(elem)
    if (bbox[1] - bbox[0]) != 0  and (bbox[3] - bbox[2]) != 0:
      ratio = min(box_dimension[2] / (bbox[1] - bbox[0]), 
                  box_dimension[3] / (bbox[3] - bbox[2]) )
    else:
      ratio = 1.0
    
    t1 = 'scale(' + str(ratio) + ')'
    m1 = simpletransform.parseTransform(t1)
    simpletransform.applyTransformToNode(m1, elem)
    
    # scale moves the center as well
    t2 = 'translate(' + \
       str( box_dimension[0] + box_dimension[2] / 2 - \
            (bbox[0] + (bbox[1] - bbox[0]) / 2 ) * ratio ) + ',' + \
       str( box_dimension[1] + box_dimension[3] / 2 - \
            (bbox[2] + (bbox[3] - bbox[2]) / 2 ) * ratio ) +  ')' 
    m2 = simpletransform.parseTransform(t2)
    simpletransform.applyTransformToNode(m2, elem)
    
  
  # Find or Create a layer given layer name and parent
  def find_create_layer(self, parent, layer_name):
    # this should not be svg:g
    path = '//g[@inkscape:label="%s"]' % layer_name
    path += '|//svg:g[@inkscape:label="%s"]' % layer_name
    el_list = self.document.xpath(path, namespaces = inkex.NSS)
    #inkex.debug(el_list)
    if el_list:
      layer = el_list[0]
      #inkex.debug('this code never execute, why?')
    else:  
      layer = inkex.etree.SubElement(parent, 'g')
      layer.set(inkex.addNS('label', 'inkscape'), layer_name )
      layer.set(inkex.addNS('groupmode', 'inkscape'), 'layer')
    return layer
    
  def create_group(self, group_name):
    group = inkex.etree.Element('g')
    group.set(inkex.addNS('label', 'inkscape'), group_name)
    group.set('fill', 'none')
    return group


if __name__ == '__main__':
  e = Merge_Svgs()
  e.affect()
